/*-
 * Copyright (C)2010..2016 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 * Used tcpmssd (c) 2000 by Ruslan Ermilov as reference.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2010..2016 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: mssseparated.c,v 1.6 2016/03/23 20:43:37 babolo Exp $";
#endif /* not lint */

#define BLIN_COMPAT  3
#define Bpars_COMPAT 3
#define MIFE_COMPAT  4

#include <sys/types.h>
#include <sys/event.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sysexits.h>
#include <strings.h>
#include <unistd.h>
#include <limits.h>
#include <signal.h>
#include <stdlib.h>
#include <netdb.h>
#include <errno.h>
#include <stdio.h>
#include <time.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <mife.h>

#define MSSSEPARATE_FLOG  "/var/log/mssseparated"
#define MSSSEPARATE_FPID  "/var/run/mssseparated.pid"
#define MSSSEPARATE_FCONF "/usr/local/etc/mssseparated.conf"
#define MSSSEPARATE_MIIN  (-8)

#define MSSSEPARATE_SLEEP  0x0100 /*             */
#define MSSSEPARATE_CFGENV 0x0200 /*     */
#define MSSSEPARATE_DAEMON 0x0400 /*                           */

typedef struct {
    int32_t min;
    int32_t max;
    int32_t rule;
} mssseparated_icfgentry;

typedef struct {
    int32_t top;
    mssseparated_icfgentry entry[1];
} mssseparated_icfg;

char sterm, shup;

static void
sigterm_handler(int sig) {
    sterm = 1;
}

static void
sighup_handler(int sig) {
    shup = 1;
}

static void
usage(int ex) {
    fprintf( stderr
           , "mssseparated @BABOLO V.M "VERS"  "DATE" http://www.babolo.ru/\n"
             "   Separates TCP SYN packets on ipfw divert socket by MSS\n"
             "\n"
             "Usage:\n"
             "mssseparated [-qv[v..]] [-f conffile] [-p pidfile] -d port\n"
             "   conffile %s by default (-F)\n"
             "   pidfile %s by default (-P)\n"
           , MSSSEPARATE_FCONF
           , MSSSEPARATE_FPID
           )
    ;
    exit(ex);
}

static void
printcfg(u_int32_t flags, mssseparated_icfg *icfg) {
#   define blin_internal_flags (flags)
    ssize_t i;

    if  (!icfg) {
        ifBLIN_QX0("Config absent");
        return;
    }
    fprintf(stderr, "# top->%d\n", icfg->top);
    for (i = 0; i < icfg->top; i++) {
        if  (icfg->entry[i].min == MSSSEPARATE_MIIN) {
            if  (icfg->entry[i].max == 0xFFFF) {
            ;   if  (icfg->entry[i].rule < 0) {
            ;       fprintf(stderr, "                  # %2d\n", i);
            ;   } else {
            ;       fprintf(stderr, "            %5d # %2d\n", icfg->entry[i].rule, i);
            ;   }
            } else if (icfg->entry[i].min == icfg->entry[i].max) {
            ;   if  (icfg->entry[i].rule < 0) {
            ;       fprintf(stderr, "%d                # %2d\n", icfg->entry[i].min, i);
            ;   } else {
            ;       fprintf( stderr, "%d          %5d # %2d\n"
                           , icfg->entry[i].min, icfg->entry[i].rule, i
                           );
            ;   }
            } else {
            ;   if  (icfg->entry[i].rule < 0) {
            ;       fprintf( stderr, "%d-%d   %s       # %2d\n"
                           , icfg->entry[i].min, icfg->entry[i].max
                           , ( (icfg->entry[i].max < 100)
                             ? ( (icfg->entry[i].max < 10)
                               ? ((icfg->entry[i].max < 0) ? "   " : "    ")
                               : "   "
                               )
                             : ( (icfg->entry[i].max < 10000)
                               ? ((icfg->entry[i].max < 1000) ? "  " : " ")
                               : ""
                             ) )
                           , i
                           );
            ;   } else {
            ;       fprintf( stderr, "%d-%d   %s %5d # %2d\n"
                           , icfg->entry[i].min, icfg->entry[i].max
                           , ( (icfg->entry[i].max < 100)
                             ? ( (icfg->entry[i].max < 10)
                               ? ((icfg->entry[i].max < 0) ? "   " : "    ")
                               : "   "
                               )
                             : ( (icfg->entry[i].max < 10000)
                               ? ((icfg->entry[i].max < 1000) ? "  " : " ")
                               : ""
                             ) )
                           , icfg->entry[i].rule
                           , i
                           );
            }   }
        } else {
            if  (icfg->entry[i].min == icfg->entry[i].max) {
            ;   if  (icfg->entry[i].rule < 0) {
            ;       fprintf( stderr, "%d      %s       # %2d\n"
                           , icfg->entry[i].min
                           , ( (icfg->entry[i].min < 100)
                             ? ( (icfg->entry[i].min < 10)
                               ? ((icfg->entry[i].min < 0) ? "   " : "    ")
                               : "   "
                               )
                             : ( (icfg->entry[i].min < 10000)
                               ? ((icfg->entry[i].min < 1000) ? "  " : " ")
                               : ""
                             ) )
                           , i
                           );
            ;   } else {
            ;       fprintf( stderr, "%d      %s %5d # %2d\n"
                           , icfg->entry[i].min
                           , ( (icfg->entry[i].min < 100)
                             ? ( (icfg->entry[i].min < 10)
                               ? ((icfg->entry[i].min < 0) ? "   " : "    ")
                               : "   "
                               )
                             : ( (icfg->entry[i].min < 10000)
                               ? ((icfg->entry[i].min < 1000) ? "  " : " ")
                               : ""
                             ) )
                           , icfg->entry[i].rule
                           , i
                           );
            ;   }
            } else {
            ;   if  (icfg->entry[i].rule < 0) {
            ;       fprintf( stderr, "%d-%d%s%s       # %2d\n"
                           , icfg->entry[i].min, icfg->entry[i].max
                           , ( (icfg->entry[i].min < 100)
                             ? ( (icfg->entry[i].min < 10)
                               ? ((icfg->entry[i].min < 0) ? "   " : "    ")
                               : "   "
                               )
                             : ( (icfg->entry[i].min < 10000)
                               ? ((icfg->entry[i].min < 1000) ? "  " : " ")
                               : ""
                             ) )
                           , ( (icfg->entry[i].max < 100)
                             ? ( (icfg->entry[i].max < 10)
                               ? ((icfg->entry[i].max < 0) ? "   " : "    ")
                               : "   "
                               )
                             : ( (icfg->entry[i].max < 10000)
                               ? ((icfg->entry[i].max < 1000) ? "  " : " ")
                               : ""
                             ) )
                           , i
                           );
            ;   } else {
            ;       fprintf( stderr, "%d-%d%s%s %5d # %2d\n"
                           , icfg->entry[i].min, icfg->entry[i].max
                           , ( (icfg->entry[i].min < 100)
                             ? ( (icfg->entry[i].min < 10)
                               ? ((icfg->entry[i].min < 0) ? "   " : "    ")
                               : "   "
                               )
                             : ( (icfg->entry[i].min < 10000)
                               ? ((icfg->entry[i].min < 1000) ? "  " : " ")
                               : ""
                             ) )
                           , ( (icfg->entry[i].max < 100)
                             ? ( (icfg->entry[i].max < 10)
                               ? ((icfg->entry[i].max < 0) ? "   " : "    ")
                               : "   "
                               )
                             : ( (icfg->entry[i].max < 10000)
                               ? ((icfg->entry[i].max < 1000) ? "  " : " ")
                               : ""
                             ) )
                           , icfg->entry[i].rule
                           , i
                           );
}   }   }   }   }
#   undef blin_internal_flags

mssseparated_icfg *
readcfg(u_int32_t flags, const char *confi) {
#   define blin_internal_flags (flags)
    mssseparated_icfg *icfg;
    mife_descriptor *md;
    size_t ilen = 340;
    babolo_parm *bp;
    int ex = EX_OK;
    u_int32_t bpfl;
    ssize_t clen;
    char *cfg;
    char *c;

    if  (!confi || !*confi) {
        ifBLIN_QW0("Config file name absent");
        icfg = NULL;
        goto out;
    }
    if  (!(md = mife_init(MIFE_FULL))) {
        ifBLIN_QW0("mife_init");
        ex = EX_IOERR;
        goto out;
    }
    if  (0 > mife_ctl(md, MIFE_CTLFILE, confi)) {
        ifBLIN_QW0("MIFE_CTLFILE");
        ex = EX_IOERR;
        goto out;
    }
    if  (0 > (clen = mife_ctl(md, MIFE_CTLREAD, (off_t)0, (ssize_t)0))) {
        ifBLIN_QW0("MIFE_CTLREAD");
        ex = EX_IOERR;
        goto out;
    }
    if  (!(cfg = mife_point(md, MIFE_POINOFF, (off_t)0))) {
        ifBLIN_QW0("MIFE_POINOFF");
        ex = EX_IOERR;
        goto out;
    }
    icfg = malloc(ilen * sizeof(mssseparated_icfgentry) + sizeof(mssseparated_icfg));
    for (icfg->top = 0; cfg && *cfg; ) {
        if  (*cfg == '#') {
            cfg = index(cfg, '\n');
            continue;
        }
        if  (icfg->top >= ilen) {
            ilen *= 2;
            icfg = reallocf(icfg, ilen * sizeof(mssseparated_icfgentry) + sizeof(mssseparated_icfg));
            if  (!icfg) {
                ifBLIN_QX0("no RAM #2");
                goto out;
        }   }
        bpfl = Bpars_CRLN | Bpars_LFLN | Bpars_FAST | Bpars_FIXB | Bpars_NOAL | Bpars_FIRM;
        if  (!(flags & MSSSEPARATE_CFGENV)) bpfl |= Bpars_NOEN;
        bp = babolo_getparm(bpfl | 3, &cfg, "#", 0);
        if  (bp->argc <= 1 && !bp->argv[0]) continue;
        icfg->entry[icfg->top].min = MSSSEPARATE_MIIN;
        icfg->entry[icfg->top].max = 0xFFFF;
        icfg->entry[icfg->top].rule = -1;
        c = bp->argv[0];
        if  (bp->argc && *c) {
            icfg->entry[icfg->top].min = strtol(c, &c, 0);
            if  (*c == '-') {
                icfg->entry[icfg->top].max = strtol(++c, &c, 0);
            } else {
                icfg->entry[icfg->top].max = icfg->entry[icfg->top].min;
        }   }
        if  (icfg->entry[icfg->top].min < MSSSEPARATE_MIIN)
            icfg->entry[icfg->top].min = MSSSEPARATE_MIIN;
        if  (icfg->entry[icfg->top].min > 0xFFFF)
            icfg->entry[icfg->top].min = 0xFFFF;
        if  (icfg->entry[icfg->top].max < MSSSEPARATE_MIIN)
            icfg->entry[icfg->top].max = MSSSEPARATE_MIIN;
        if  (icfg->entry[icfg->top].max > 0xFFFF)
            icfg->entry[icfg->top].max = 0xFFFF;
        c = bp->argv[1];
        if  ((bp->argc > 1) && *c) {
            icfg->entry[icfg->top].rule = strtol(c, &c, 0);
        }
        if  (icfg->entry[icfg->top].rule < -1) icfg->entry[icfg->top].rule = -1;
        if  (icfg->entry[icfg->top].rule > 0xFFFF) icfg->entry[icfg->top].rule = 0xFFFF;
        icfg->top++;
    }
    if  (0 > mife_ctl(md, MIFE_CTLFINI)) {
        ifBLIN_QW1("MIFE_CTLFINI");
    }
out:
    return(icfg);
#   undef blin_internal_flags
}

static void
pidsave(u_int32_t flags, const char *pidfi, u_int16_t port) {
#   define blin_internal_flags (flags)
    if  (pidfi) {
        FILE *pidfile;
        if  (!(pidfile = fopen(pidfi, "w"))) {
            ifBLIN_QW0("fopen %s", pidfi);
        }
        fprintf(pidfile, "%d\n-p %d\n", (int)getpid(), (int)port);
        fclose(pidfile);
}   }
#   undef blin_internal_flags

static void
printlog(FILE *fout, struct ip *buf, struct tcphdr *tc, int32_t mss, in_port_t port) {
    struct protoent *proto;
    struct tm *timep;
    char timec[20];
    time_t timev;

    if  (time(&timev) == (time_t)-1) {
        timec[0] = '\0';
    } else if (!(timep = localtime(&timev))) {
        timec[0] = '\0';
    } else if (!strftime(timec, 20, "%Y-%m-%d_%H:%M:%S", timep)) {
        timec[0] = '\0';
    }
    proto = getprotobynumber(buf->ip_p);
    if  (tc) {
        fprintf( fout, "%s %s %u.%u.%u.%u:%u > %u.%u.%u.%u:%u%s%s%s%s%s%s%s%s mss=%d rule=%u\n"
               , timec
               , proto->p_name
               ,  buf->ip_src.s_addr        & 0xFF
               , (buf->ip_src.s_addr >> 8 ) & 0xFF
               , (buf->ip_src.s_addr >> 16) & 0xFF
               , (buf->ip_src.s_addr >> 24) & 0xFF
               , ntohs(tc->th_sport)
               ,  buf->ip_dst.s_addr        & 0xFF
               , (buf->ip_dst.s_addr >> 8 ) & 0xFF
               , (buf->ip_dst.s_addr >> 16) & 0xFF
               , (buf->ip_dst.s_addr >> 24) & 0xFF
               , ntohs(tc->th_dport)
               , (tc->th_flags & TH_FIN) ? " fin" : ""
               , (tc->th_flags & TH_SYN) ? " syn" : ""
               , (tc->th_flags & TH_RST) ? " rst" : ""
               , (tc->th_flags & TH_PUSH) ? " push" : ""
               , (tc->th_flags & TH_ACK) ? " ack" : ""
               , (tc->th_flags & TH_URG) ? " urg" : ""
               , (tc->th_flags & TH_ECE) ? " ece" : ""
               , (tc->th_flags & TH_CWR) ? " cwr" : ""
               , mss
               , port
               );
    } else {
        fprintf( fout, "%s %s %u.%u.%u.%u->%u.%u.%u.%u nomss rule=%u\n"
               , timec
               , proto->p_name
               ,  buf->ip_src.s_addr        & 0xFF
               , (buf->ip_src.s_addr >> 8 ) & 0xFF
               , (buf->ip_src.s_addr >> 16) & 0xFF
               , (buf->ip_src.s_addr >> 24) & 0xFF
               ,  buf->ip_dst.s_addr        & 0xFF
               , (buf->ip_dst.s_addr >> 8 ) & 0xFF
               , (buf->ip_dst.s_addr >> 16) & 0xFF
               , (buf->ip_dst.s_addr >> 24) & 0xFF
               , port
               );
}   }

static char *
rediropt(u_int32_t flags, char *optarg, int redirenv) {
    char *cr;

    if  (redirenv > 0) {
        if  (!(cr = getenv(optarg))) {
            ifBLIN_QX0("No ENV %s", optarg);
            exit(EX_USAGE);
        }
    }else{
        cr = optarg;
    }
    return(cr);
}

#define REEND(E)                      \
    if  (flags & MSSSEPARATE_SLEEP) { \
        sleep(seconds);               \
        goto in;                      \
    } else {                          \
        ex = E;                       \
        goto out;                     \
    }

int
main(int argc, char **argv) {
    const char *confi, *pidfi, *logfi;
    mssseparated_icfg *icfg;
    struct sockaddr_in sin;
    struct ip *buf = NULL;
    unsigned int seconds;
    struct kevent kev;
    int redirenv, ex;
    babolo_parm *bp;
    babolo_opts *bos;
    u_int32_t flags;
    u_int16_t port;
    FILE *logg;
    char *o, c;
    int s, kq;

    pidfi = NULL;
    confi = NULL;
    logfi = NULL;
    logg = NULL;
    port = 0;
    s = -1;
in:
    port = 0;
    ex = EX_OK;
    seconds = 0;
    pidfi = NULL;
    blin_ctl(BLIN_CTL_FSET, (flags = BLIN_GEN0));
#   define blin_internal_flags (flags)
    if  (!(bos = babolo_openopts(BLIN_GEN0 | Bpars_NONU | Bpars_SYME | Bpars_ESYM, 'Z'))) {
        ifBLIN_QW0("babolo_openopts");
        ex = EX_SOFTWARE;
        goto out;
    }
    if  (babolo_setopts(bos, Bpars_NONU, argc, argv, "Dd:eFf:Ll:Pp:qs:v?")) {
        ifBLIN_QW0("babolo_setopts");
        ex = EX_SOFTWARE;
        goto out;
    }
    for (; 0 < (c = babolo_getopts(bos));) {
        switch (c) {
        case 'D': flags |= MSSSEPARATE_DAEMON;
                ; break;
        case 'd': port = babolo_getoptunum(bos, 0);
                ; if  (!!errno) {
                      ifBLIN_QW0("port N absent");
                      ex = EX_USAGE;
                      goto out;
                  }
                ; break;
        case 'e': flags |= MSSSEPARATE_CFGENV;
                ; break;
        case 'F': confi = MSSSEPARATE_FCONF;
                ; break;
        case 'f': confi = babolo_getoptsarg(bos);
                ; if  (!confi) {
                      ifBLIN_QX0("conf file name absent");
                      ex = EX_USAGE;
                      goto out;
                  }
                ; break;
        case 'L': logfi = MSSSEPARATE_FLOG;
                ; break;
        case 'l': logfi = babolo_getoptsarg(bos);
                ; if  (!logfi) {
                      ifBLIN_QX0("log file name absent");
                      ex = EX_USAGE;
                      goto out;
                  }
                ; break;
        case 'P': pidfi = MSSSEPARATE_FPID;
                ; break;
        case 'p': pidfi = babolo_getoptsarg(bos);
                ; if  (!pidfi) {
                      ifBLIN_QX0("PID file name absent");
                      ex = EX_USAGE;
                      goto out;
                  }
                ; break;
        case 'q': flags &= ~BLIN_MASK;
                ; break;
        case 's': seconds = babolo_getoptunum(bos, 0);
                ; if  (!!errno) {
                      ifBLIN_QX0("sleep time absent");
                      ex = EX_USAGE;
                      goto out;
                  }
                ; flags |= MSSSEPARATE_SLEEP;
                ; break;
        case 'v': BLIN_VERBOSE(flags);
                ; break;
        case '?': usage(EX_OK);
        default : usage(EX_USAGE);
        }
        if  (ex) goto out;
    }
    if  (!(icfg = readcfg(flags, confi))) {
        ex = EX_OSERR;
        goto out;
    }
    if  (logfi) blin_ctl(BLIN_CTL_FNAM | BLIN_GEN7, logfi);
    blin_ctl(BLIN_CTL_WHAT | BLIN_GEN7, BLIN_MODPRNT | BLIN_MODEOAU);
    ifBLIN_QO2 printcfg(flags, icfg);
    if  ((flags & MSSSEPARATE_DAEMON) && daemon(0, 0)) {
        ifBLIN_QW0("daemon()");
        ex = EX_OSERR;
        goto out;
    }
    if  (s >= 0) close(s);
    if  ((s = socket(PF_INET, SOCK_RAW, IPPROTO_DIVERT)) < 0) {
        ifBLIN_QW0("socket()");
        ex = EX_OSERR;
        goto out;
    }
    bzero(&sin, sizeof(sin));
    sin.sin_family = PF_INET;
    sin.sin_addr.s_addr = INADDR_ANY;
    sin.sin_port = htons(port);
    if  (bind(s, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
        ifBLIN_QW0("bind()");
        ex = EX_OSERR;
        goto out;
    }
    pidsave(flags, pidfi, port);
    sterm = shup = 0;
    if  (signal(SIGTERM, sigterm_handler) == SIG_ERR) {
        ifBLIN_QW0("signal(TERM)");
    }
    if  (signal(SIGINT, sigterm_handler) == SIG_ERR) {
        ifBLIN_QW0("signal(INT)");
    }
    if  (signal(SIGHUP, sighup_handler) == SIG_ERR) {
        ifBLIN_QW0("signal(HUP)");
    }
    if  ((kq = kqueue()) < 1) {
        ifBLIN_QW0("kqueue()");
        ex = EX_OSERR;
        goto out;
    }
    EV_SET(&kev, s, EVFILT_READ, EV_ADD, 0, 0, NULL);
    if  (!buf) {
        if  (!(buf = malloc(IP_MAXPACKET))) {
            ifBLIN_QX0("no RAM #1");
            ex = EX_OSERR;
            goto out;
    }   }
    if  (kevent(kq, &kev, 1, NULL, 0, NULL) < 0) {
        ifBLIN_QW0("kevent() set");
        ex = EX_OSERR;
        goto out;
    }
    while (1) {
        u_char *opt;
        int32_t mss;
        socklen_t sinlen;
        struct tcphdr *tc;
        ssize_t plen, hlen, i;
        int tlen, olen, optlen;

        tc = NULL;
        switch(kevent(kq, NULL, 0, &kev, 1, NULL)) {
        default: if  (errno == EINTR) goto intrpt;
               ; ifBLIN_QW0("kevent() get");
               ; REEND(EX_OSERR);
               ; /* NOTREACHED */
        case 1 : mss = -1;
               ; sinlen = sizeof(sin);
               ; plen = recvfrom(s, buf, IP_MAXPACKET, 0, (struct sockaddr *)&sin, &sinlen);
               ; if  (plen < 0) {
               ;     if (errno == EINTR || errno == EAGAIN) continue;
               ;     ifBLIN_QW0("recvfrom");
               ;     REEND(EX_IOERR);
               ; }
               ; hlen = buf->ip_hl << 2;
               ; if  (  buf->ip_p == IPPROTO_TCP
                     && (ntohs(buf->ip_off) & IP_OFFMASK) == 0
                     && ntohs(buf->ip_len) == plen
                     && hlen <= plen
                     && plen - hlen >= sizeof(struct tcphdr)
                     ) {
               ;     mss = -2;
               ;     tc = (struct tcphdr *)((char*)buf + hlen);
               ;     tlen = tc->th_off << 2;
               ;     if  (tlen > (plen - hlen)) {
               ;         ifBLIN_QX1("tlen %d > plen %d - hlen %d", tlen, plen, hlen);
               ;         goto push;
               ;     }
               ;     mss = -3;
               ;     if  (!(tc->th_flags & TH_SYN)) continue;
               ;     mss = -4;
               ;     if  (tlen > sizeof(struct tcphdr)) {
               ;         for ( olen = tlen - sizeof(struct tcphdr), opt = (u_char *)(tc + 1)
                             ; olen > 0
                             ; olen -= optlen, opt += optlen
                             ) {
               ;             if  (*opt == TCPOPT_EOL) {
               ;                 mss = -5;
               ;                 break;
               ;             } else if (*opt == TCPOPT_NOP) {
               ;                 optlen = 1;
               ;             } else if ((optlen = *(opt + 1)) <= 0) {
               ;                 mss = -6;
               ;                 break;
               ;             } else if (optlen > olen) {
               ;                 mss = -7;
               ;                 break;
               ;             } else if (*opt == TCPOPT_MAXSEG) {
               ;                 mss = ntohs(*(u_int16_t *)(opt + 2));
               ;                 break;
               ; }   }   }   }
push:          ; for (i = icfg->top; i > 0; ) {
               ;     i--;
               ;     if  ((icfg->entry[i].min <= mss) && (icfg->entry[i].max >= mss)) {
               ;         if  (icfg->entry[i].rule >= 0) sin.sin_port = icfg->entry[i].rule;
               ;         break;
               ; }   }
               ; ifBLIN_QO2 printlog(stderr, buf, tc, mss, sin.sin_port);
               ; if  (logg) printlog(logg, buf, tc, mss, sin.sin_port);
               ; if  (sendto(s, buf, plen, 0, (struct sockaddr *)&sin, sinlen) < 0) {
               ;     if (errno == EAGAIN) continue;
               ;     ifBLIN_QW0("sendto");
               ;     REEND(EX_IOERR);
               ; };
intrpt:        ; /* FALLTHROUGH */
        case 0 : if  (sterm) {
               ;     if  (pidfi) unlink(pidfi);
               ;     sterm = 0;
               ;     exit(0);
               ; }
               ; if  (shup) {
               ;     if  (!(icfg = readcfg(flags, confi))) {
               ;         ex = EX_OSERR;
               ;         REEND(EX_IOERR);
               ;     }
               ;     if  (logg) fclose(logg);
               ;     if  (logfi) {
               ;         if  (!(logg = fopen(logfi, "a"))) {
               ;             ifBLIN_QW0("fopen %s", logfi);
               ;         }
               ;         if  (setvbuf(logg, NULL, _IONBF, 0)) {
               ;             ifBLIN_QW0("setvbuf()");
               ;     }   }
               ;     pidsave(flags, pidfi, port);
               ;     shup = 0;
               ; }
               ; continue;
    }   }
out: exit(ex);
#   undef blin_internal_flags
}
